# ******************************************************************************
#
# forms.py:  allauth_2f2a forms
#
# SPDX-License-Identifier: Apache-2.0
#
# django-allauth-2f2a, a 2fa adapter for django-allauth.
#
# ******************************************************************************
#
# Copyright 2016-2021 Víðir Valberg Guðmundsson and Percipient
# Networks, LLC.
# Copyright 2021 Jeremy A Gray <gray@flyquackswim.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.
#
# ******************************************************************************
#
"""allauth_2f2a forms."""

from django import forms
from django.utils.translation import gettext_lazy as _
from django_otp.forms import OTPAuthenticationFormMixin
from django_otp.plugins.otp_totp.models import TOTPDevice


class TOTPAuthenticateForm(OTPAuthenticationFormMixin, forms.Form):
    """TOTP authentication form."""

    otp_token = forms.CharField(
        label=_("Token"),
    )

    def __init__(self, user, **kwargs):
        """Initialize the TOTP authentication form."""
        super(TOTPAuthenticateForm, self).__init__(**kwargs)
        self.fields["otp_token"].widget.attrs.update(
            {
                "autofocus": "autofocus",
                "autocomplete": "off",
                "inputmode": "numeric",
            }
        )
        self.user = user

    def clean(self):
        """Clean TOTP authentication form data."""
        self.clean_otp(self.user)
        return self.cleaned_data


class TOTPDeviceForm(forms.Form):
    """TOTP device form."""

    token = forms.CharField(
        label=_("Token"),
    )

    def __init__(self, user, metadata=None, **kwargs):
        """Initialize the TOTP device form."""
        super(TOTPDeviceForm, self).__init__(**kwargs)
        self.fields["token"].widget.attrs.update(
            {
                "autofocus": "autofocus",
                "autocomplete": "off",
            }
        )
        self.user = user
        self.metadata = metadata or {}

    def clean_token(self):
        """Clean the TOTP device token."""
        token = self.cleaned_data.get("token")

        # Find the unconfirmed device and attempt to verify the token.
        self.device = self.user.totpdevice_set.filter(confirmed=False).first()
        if not self.device.verify_token(token):
            raise forms.ValidationError(_("The entered token is not valid"))

        return token

    def save(self):
        """Save the confirmed TOTP device."""
        # The device was found to be valid, delete other confirmed
        # devices and confirm the new device.
        self.user.totpdevice_set.filter(confirmed=True).delete()
        self.device.confirmed = True
        self.device.save()

        return self.device


class TOTPDeviceRemoveForm(forms.Form):
    """TOTP device removal form."""

    def __init__(self, user, **kwargs):
        """Initialize the TOTP device removal form."""
        super(TOTPDeviceRemoveForm, self).__init__(**kwargs)
        self.user = user

    def save(self):
        """Delete the removed TOTP device.

        Delete the removed TOTP device and remove all of its backup
        tokens.
        """
        # Delete any backup tokens.
        static_device = self.user.staticdevice_set.get(name="backup")
        static_device.token_set.all().delete()
        static_device.delete()

        # Delete TOTP device.
        device = TOTPDevice.objects.get(user=self.user)
        device.delete()



# ******************************************************************************
#
# urls.py:  URL routing for allauth_2f2a module
#
# SPDX-License-Identifier: Apache-2.0
#
# django-allauth-2f2a, a 2fa adapter for django-allauth.
#
# ******************************************************************************
#
# Copyright 2016-2021 Víðir Valberg Guðmundsson and Percipient
# Networks, LLC.
# Copyright 2021 Jeremy A Gray <gray@flyquackswim.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.
#
# ******************************************************************************
#
"""URL routing for allauth_2f2a module."""

from django.urls import re_path

from allauth_2f2a import views

urlpatterns = [
    re_path(
        r"^2fa/authenticate/?$",
        views.TwoFactorAuthenticate.as_view(),
        name="two-factor-authenticate",
    ),
    re_path(r"^2fa/setup/?$", views.TwoFactorSetup.as_view(), name="two-factor-setup"),
    re_path(
        r"^2fa/backup/?$",
        views.TwoFactorBackupTokens.as_view(),
        name="two-factor-backup",
    ),
    re_path(
        r"^2fa/remove/?$",
        views.TwoFactorRemove.as_view(),
        name="two-factor-remove",
    ),
]



# ******************************************************************************
#
# mixins.py:  allauth_2f2a mixins
#
# SPDX-License-Identifier: Apache-2.0
#
# django-allauth-2f2a, a 2fa adapter for django-allauth.
#
# ******************************************************************************
#
# Copyright 2016-2021 Víðir Valberg Guðmundsson and Percipient
# Networks, LLC.
# Copyright 2021 Jeremy A Gray <gray@flyquackswim.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.
#
# ******************************************************************************
#
"""allauth_2f2a mixins."""

from django.contrib.auth.mixins import AccessMixin
from django.http import HttpResponseRedirect
from django.urls import reverse_lazy

from allauth_2f2a.utils import user_has_valid_totp_device


class ValidTOTPDeviceRequiredMixin(AccessMixin):
    """Require a valid TOTP device."""

    no_valid_totp_device_url = reverse_lazy("two-factor-setup")

    def dispatch(self, request, *args, **kwargs):
        """Dispatch appropriate view based on device settings."""
        if not request.user.is_authenticated:
            return self.handle_no_permission()
        if not user_has_valid_totp_device(request.user):
            return self.handle_missing_totp_device()
        return super(ValidTOTPDeviceRequiredMixin, self).dispatch(
            request, *args, **kwargs
        )

    def handle_missing_totp_device(self):
        """Handle missing device.

        Redirect to ``self.no_valid_totp_device_url`` if there is not
        valid TOTP device configured.

        Returns
        -------
        django.http.HttpResponseRedirect
            A redirect to ``self.no_valid_totp_device_url``.
        """
        return HttpResponseRedirect(self.no_valid_totp_device_url)



# ******************************************************************************
#
# views.py:  allauth_2f2a views
#
# SPDX-License-Identifier: Apache-2.0
#
# django-allauth-2f2a, a 2fa adapter for django-allauth.
#
# ******************************************************************************
#
# Copyright 2016-2021 Víðir Valberg Guðmundsson and Percipient
# Networks, LLC.
# Copyright 2021 Jeremy A Gray <gray@flyquackswim.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.
#
# ******************************************************************************
#
"""allauth_2f2a views."""

import uuid
from base64 import b64encode
from pathlib import Path

from allauth.account import signals
from allauth.account.adapter import get_adapter
from allauth.account.utils import get_login_redirect_url
from django.conf import settings
from django.contrib import messages
from django.contrib.auth import get_user_model
from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponseRedirect
from django.shortcuts import redirect
from django.urls import reverse
from django.urls import reverse_lazy
from django.utils.encoding import force_str
from django.views.generic import FormView
from django.views.generic import TemplateView
from django_otp.plugins.otp_static.models import StaticToken
from django_otp.plugins.otp_totp.models import TOTPDevice

from allauth_2f2a import app_settings
from allauth_2f2a.forms import TOTPAuthenticateForm
from allauth_2f2a.forms import TOTPDeviceForm
from allauth_2f2a.forms import TOTPDeviceRemoveForm
from allauth_2f2a.mixins import ValidTOTPDeviceRequiredMixin
from allauth_2f2a.utils import generate_totp_config_svg_for_device
from allauth_2f2a.utils import user_has_valid_totp_device

from .utils import get_form_class


class TwoFactorAuthenticate(FormView):
    """2fa authentication view."""

    template_name = "allauth_2f2a/authenticate." + app_settings.TEMPLATE_EXTENSION
    form_class = TOTPAuthenticateForm

    def dispatch(self, request, *args, **kwargs):
        """Dispatch to correct 2fa authentication view."""
        # Redirect to login if not ready for 2fa.
        if "allauth_2f2a_user_id" not in request.session:
            # Don't use the redirect_to_login here since we don't actually want
            # to include the next parameter.
            return redirect("account_login")

        # Dispatch 2fa authentication.
        return super(TwoFactorAuthenticate, self).dispatch(request, *args, **kwargs)

    def get_form_class(self):
        """Get the authentication form class."""
        return get_form_class(
            app_settings.TWOFA_FORMS,
            "authentication",
            "allauth_2f2a.forms.TOTPAuthenticateForm",
        )

    def get_form_kwargs(self):
        """Grab user and insert into ``kwargs``."""
        kwargs = super(TwoFactorAuthenticate, self).get_form_kwargs()
        user_id = self.request.session["allauth_2f2a_user_id"]
        kwargs["user"] = get_user_model().objects.get(id=user_id)
        return kwargs

    def form_valid(self, form):
        """Complete the django-allauth login process.

        The allauth 2fa login flow is now done (the user logged in
        successfully with 2FA), continue the logic from
        allauth.account.utils.perform_login since it was interrupted
        earlier.
        """
        adapter = get_adapter(self.request)

        # Skip over the (already done) 2fa login flow and continue the
        # original allauth login flow.
        super(adapter.__class__, adapter).login(self.request, form.user)

        # Perform the rest of allauth.account.utils.perform_login,
        # this is copied from commit
        # cedad9f156a8c78bfbe43a0b3a723c1a0b840dbd.

        # TODO Support redirect_url.
        response = HttpResponseRedirect(get_login_redirect_url(self.request))

        # TODO Support signal_kwargs.
        signals.user_logged_in.send(
            sender=form.user.__class__,
            request=self.request,
            response=response,
            user=form.user,
        )

        adapter.add_message(
            self.request,
            messages.SUCCESS,
            "account/messages/logged_in.txt",
            {"user": form.user},
        )

        return response


class TwoFactorSetup(LoginRequiredMixin, FormView):
    """2fa setup view."""

    template_name = "allauth_2f2a/setup." + app_settings.TEMPLATE_EXTENSION
    form_class = TOTPDeviceForm
    success_url = reverse_lazy("two-factor-backup")

    def dispatch(self, request, *args, **kwargs):
        """Dispatch to setup or backup view."""
        # If the user has 2fa configured, redirect them to the backup
        # tokens.
        if user_has_valid_totp_device(request.user):
            return HttpResponseRedirect(reverse("two-factor-backup"))

        # Continue to 2fa setup otherwise.
        return super(TwoFactorSetup, self).dispatch(request, *args, **kwargs)

    def get_form_class(self):
        """Get the device creation form class."""
        return get_form_class(
            app_settings.TWOFA_FORMS,
            "device",
            "allauth_2f2a.forms.TOTPDeviceForm",
        )

    def _new_device(self):
        """Generate a new ``django_otp.plugins.otp_totp.models.TOTPDevice``.

        Replace any unconfirmed TOTPDevices with a new one for
        confirmation.

        This needs to be done whenever a GET request to the page is
        received OR if the confirmation of the device fails.
        """
        self.request.user.totpdevice_set.filter(confirmed=False).delete()
        self.device = TOTPDevice.objects.create(user=self.request.user, confirmed=False)

    def get(self, request, *args, **kwargs):
        """Generate a new TOTP device on every request."""
        self._new_device()
        return super(TwoFactorSetup, self).get(request, *args, **kwargs)

    def get_qr_code_data_uri(self):
        """Generate QR code image data.

        Generate the 2FA QR code image data using either the 'data:'
        protocol with a base64 encoded string or using a locally
        generated file with a random filename.

        Returns
        -------
        string
            The URI of the image; either a base64 encoded SVG image
            string with the 'data:' protocol or the URL of the SVG
            image file.  Use of the 'data:' protocol requires a CSP
            that allows that protocol for images.

        Raises
        ------
        ImproperlyConfigured
            If ``settings.MEDIA_ROOT/qrcodes`` is not a directory.
        """
        # Generate the QR code image.
        svg_data = generate_totp_config_svg_for_device(self.request, self.device)

        # Serve QR code from file.
        if app_settings.QRCODE_TYPE == "file":
            qr_dir = Path(settings.MEDIA_ROOT) / "qrcodes"
            # Raise exception if directory does not exist.
            if not qr_dir.is_dir():
                raise ImproperlyConfigured

            # Generate a UUID file name to prevent prediction and to
            # generate a useable Path() and URI.
            id = uuid.uuid4().hex
            fn = qr_dir / (id + ".svg")
            uri = "/media/qrcodes/" + id + ".svg"

            # svg_data is in bytes; write it to specified file.
            with open(fn, "wb") as f:
                f.write(svg_data)
            return uri
        # Serve QR code from data: protocol.  Beware the CSP implications.
        else:
            return f"data:image/svg+xml;base64,{force_str(b64encode(svg_data))}"

    def get_context_data(self, **kwargs):
        """Add ``qr_code_url`` to the context."""
        context = super(TwoFactorSetup, self).get_context_data(**kwargs)
        context["qr_code_url"] = self.get_qr_code_data_uri()
        return context

    def get_form_kwargs(self):
        """Add the user to the form ``kwargs``."""
        kwargs = super(TwoFactorSetup, self).get_form_kwargs()
        kwargs["user"] = self.request.user
        return kwargs

    def form_valid(self, form):
        """Save the device and continue form processing."""
        form.save()
        return super(TwoFactorSetup, self).form_valid(form)

    def form_invalid(self, form):
        """Generate a new device for an invalid form."""
        self._new_device()
        return super(TwoFactorSetup, self).form_invalid(form)


class TwoFactorRemove(ValidTOTPDeviceRequiredMixin, FormView):
    """2fa device removal."""

    template_name = "allauth_2f2a/remove." + app_settings.TEMPLATE_EXTENSION
    form_class = TOTPDeviceRemoveForm
    success_url = reverse_lazy("two-factor-setup")

    def get_form_class(self):
        """Get the device removal form class."""
        return get_form_class(
            app_settings.TWOFA_FORMS,
            "remove",
            "allauth_2f2a.forms.TOTPDeviceRemoveForm",
        )

    def form_valid(self, form):
        """Save the removal data and continue form processing."""
        form.save()
        return super(TwoFactorRemove, self).form_valid(form)

    def get_form_kwargs(self):
        """Add the user to the form ``kwargs``."""
        kwargs = super(TwoFactorRemove, self).get_form_kwargs()
        kwargs["user"] = self.request.user
        return kwargs


class TwoFactorBackupTokens(ValidTOTPDeviceRequiredMixin, TemplateView):
    """2fa backup token generation."""

    template_name = "allauth_2f2a/backup_tokens." + app_settings.TEMPLATE_EXTENSION
    # This can be overridden in a subclass to ``True``, to have that
    # particular view always reveal the tokens.
    reveal_tokens = bool(app_settings.ALWAYS_REVEAL_BACKUP_TOKENS)

    def get_context_data(self, **kwargs):
        """Add token information to the context.

        Adds ``backup_tokens`` and ``reveal_tokens`` to the view
        context.

        Returns
        -------
        dict
            The context dictionary.
        """
        context = super(TwoFactorBackupTokens, self).get_context_data(**kwargs)
        static_device, _ = self.request.user.staticdevice_set.get_or_create(
            name="backup"
        )

        if static_device:
            context["backup_tokens"] = static_device.token_set.all()
            context["reveal_tokens"] = self.reveal_tokens

        return context

    def post(self, request, *args, **kwargs):
        """Handle 2fa backup POST."""
        static_device, _ = request.user.staticdevice_set.get_or_create(name="backup")
        static_device.token_set.all().delete()
        for _ in range(3):
            static_device.token_set.create(token=StaticToken.random_token())
        self.reveal_tokens = True
        return self.get(request, *args, **kwargs)



# ******************************************************************************
#
# utils.py:  utilities for allauth_2f2a
#
# SPDX-License-Identifier: Apache-2.0
#
# django-allauth-2f2a, a 2fa adapter for django-allauth.
#
# ******************************************************************************
#
# Copyright 2016-2021 Víðir Valberg Guðmundsson and Percipient
# Networks, LLC.
# Copyright 2021 Jeremy A Gray <gray@flyquackswim.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.
#
# ******************************************************************************
#
"""Utilities for allauth_2f2a."""

import importlib
from base64 import b32encode
from io import BytesIO
from urllib.parse import quote
from urllib.parse import urlencode

import qrcode
from django.contrib.sites.shortcuts import get_current_site
from qrcode.image.svg import SvgPathImage


def generate_totp_config_svg(device, issuer, label):
    """Generate a TOTP configuration SVG.

    Returns
    -------
    io.BytesIO
        SVG image as a memory-based file.
    """
    params = {
        "secret": b32encode(device.bin_key).decode("utf-8"),
        "algorithm": "SHA1",
        "digits": device.digits,
        "period": device.step,
        "issuer": issuer,
    }

    otpauth_url = "otpauth://totp/{label}?{query}".format(
        label=quote(label),
        query=urlencode(params),
    )

    img = qrcode.make(otpauth_url, image_factory=SvgPathImage)
    io = BytesIO()
    img.save(io)

    return io.getvalue()


def generate_totp_config_svg_for_device(request, device):
    """Generate a TOTP configuration SVG for a device.

    Returns
    -------
    io.BytesIO
        SVG image as a memory-based file.
    """
    issuer = get_current_site(request).name
    label = "{issuer}: {username}".format(
        issuer=issuer,
        username=request.user.get_username(),
    )

    return generate_totp_config_svg(
        device=device,
        issuer=issuer,
        label=label,
    )


def user_has_valid_totp_device(user):
    """Determine if the user has a valid TOTP device.

    Parameters
    ----------
    user
        The current Django user object.

    Returns
    -------
    boolean
        ``True`` if ``user`` has a valid TOTP device, ``False``
        otherwise.
    """
    if not user.is_authenticated:
        return False

    return user.totpdevice_set.filter(confirmed=True).exists()


# Idea and code borrowed from django-alluth form configuration.
def import_attribute(path):
    """Import an attribute given its string path."""
    assert isinstance(path, str)
    pkg, attr = path.rsplit(".", 1)
    ret = getattr(importlib.import_module(pkg), attr)
    return ret


def get_form_class(forms, form_id, default_form):
    """Return a form class."""
    form_class = forms.get(form_id, default_form)
    if isinstance(form_class, str):
        form_class = import_attribute(form_class)
    return form_class



# ******************************************************************************
#
# adapter.py:  django-allauth adapter for 2fa
#
# SPDX-License-Identifier: Apache-2.0
#
# django-allauth-2f2a, a 2fa adapter for django-allauth.
#
# ******************************************************************************
#
# Copyright 2016-2021 Víðir Valberg Guðmundsson and Percipient
# Networks, LLC.
# Copyright 2021 Jeremy A Gray <gray@flyquackswim.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.
#
# ******************************************************************************
#
"""Django-allauth adapter for 2fa."""

from urllib.parse import urlencode

from allauth.account.adapter import DefaultAccountAdapter
from allauth.exceptions import ImmediateHttpResponse
from django.http import HttpResponseRedirect
from django.urls import reverse

from allauth_2f2a.utils import user_has_valid_totp_device


class OTPAdapter(DefaultAccountAdapter):
    """Django-allauth OTP adapter."""

    def has_2fa_enabled(self, user):
        """Determine if the user has 2fa configured.

        Returns
        -------
        boolean
            ``True`` if the user has 2fa configured, ``False``
            otherwise.
        """
        return user_has_valid_totp_device(user)

    def login(self, request, user):
        """Require 2fa for login if configured.

        Require 2fa for login if it has been configured by the user.

        Returns
        -------
        allauth.account.adapter.DefaultAccountAdapter
            The original django-allauth account adapter if 2fa is not
            enabled.

        Raises
        ------
        allauth.exceptions.ImmediateHttpResponse
            Redirects to 2fa URL if enabled.
        """
        if self.has_2fa_enabled(user):
            # Cast to string for the case when this is not a JSON
            # serializable object.
            request.session["allauth_2f2a_user_id"] = str(user.id)

            redirect_url = reverse("two-factor-authenticate")
            # Add "next" parameter to the URL.
            view = request.resolver_match.func.view_class()
            view.request = request
            success_url = view.get_success_url()
            query_params = request.GET.copy()
            if success_url:
                query_params[view.redirect_field_name] = success_url
            if query_params:
                redirect_url += "?" + urlencode(query_params)

            raise ImmediateHttpResponse(
                response=HttpResponseRedirect(redirect_url),
            )

        # Otherwise defer to the original allauth adapter.
        return super(OTPAdapter, self).login(request, user)



# ******************************************************************************
#
# middleware.py:  allauth_2f2a middleware
#
# SPDX-License-Identifier: Apache-2.0
#
# django-allauth-2f2a, a 2fa adapter for django-allauth.
#
# ******************************************************************************
#
# Copyright 2016-2021 Víðir Valberg Guðmundsson and Percipient
# Networks, LLC.
# Copyright 2021 Jeremy A Gray <gray@flyquackswim.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.
#
# ******************************************************************************
#
"""allauth_2f2a middleware."""

import warnings

from allauth.account.adapter import get_adapter
from django.conf import settings
from django.contrib import messages
from django.shortcuts import redirect
from django.urls import NoReverseMatch
from django.urls import resolve
from django.urls import reverse
from django.utils.deprecation import MiddlewareMixin


class AllauthTwoFactorMiddleware(MiddlewareMixin):
    """Prevent partially authenticated users.

    Reset the login flow if another page is loaded halfway through the
    login.  (I.e. if the user has logged in with a username/password,
    but not yet entered their two-factor credentials.) This makes sure
    a user does not stay half logged in by mistake.

    """

    def process_request(self, request):
        """Ensure 2FA completion.

        Remove ``allauth_2f2a_user_id`` from session if the URL does
        not match the 2FA URL.
        """
        match = resolve(request.path)
        if not match.url_name or not match.url_name.startswith(
            "two-factor-authenticate"
        ):
            try:
                del request.session["allauth_2f2a_user_id"]
            except KeyError:
                pass


class BaseRequire2FAMiddleware(MiddlewareMixin):
    """Require users to configure 2FA.

    Ensure that particular users have two-factor authentication
    enabled before they have access to the rest of the app.

    If they don't have 2FA enabled, they will be redirected to the 2FA
    enrollment page and not be allowed to access other pages.
    """

    # List of URLs that the user should still be allowed to access.
    allowed_pages = [
        # Users should be able to log out or change password.
        "account_logout",
        "account_change_password",
        "account_reset_password",
        # Users should be able to configure 2FA.
        "two-factor-setup",
    ]

    # The message to the user if they don't have 2FA enabled and must
    # enable it.
    require_2fa_message = (
        "You must enable two-factor authentication before doing anything else."
    )

    def on_require_2fa(self, request):
        """Redirect to 2fa setup if required.

        If the current request requires 2fa and the user does not have
        it enabled, this is executed. The result of this is returned
        from the middleware.
        """
        # See allauth.account.adapter.DefaultAccountAdapter.add_message.
        if "django.contrib.messages" in settings.INSTALLED_APPS:
            # If there is already a pending message related to two-factor (likely
            # created by a redirect view), simply update the message text.
            storage = messages.get_messages(request)
            tag = "2fa_required"
            for m in storage:
                if m.extra_tags == tag:
                    m.message = self.require_2fa_message
                    break
            # Otherwise, create a new message.
            else:
                messages.error(request, self.require_2fa_message, extra_tags=tag)
            # Mark the storage as not processed so they'll be shown to the user.
            storage.used = False

        # Redirect user to two-factor setup page.
        return redirect("two-factor-setup")

    def require_2fa(self, request):
        """Determine if 2fa is required.

        Check if this request is required to have 2FA before accessing
        the app.

        This should return True if this request requires 2FA. (Note
        that the user was already)

        You can access anything on the request, but generally
        request.user will be most interesting here.
        """
        raise NotImplementedError("You must implement require_2fa.")

    def process_view(self, request, view_func, view_args, view_kwargs):
        """Process view based on 2fa requirements."""
        # The user is not logged in, do nothing.
        if request.user.is_anonymous:
            return

        # If this doesn't require 2FA, then stop processing.
        if not self.require_2fa(request):
            return

        # If the user is on one of the allowed pages, do nothing.
        for urlname in self.allowed_pages:
            try:
                if request.path == reverse(urlname):
                    return
            except NoReverseMatch:
                # The developer may have misconfigured the list of
                # allowed pages.  Let's not outright crash at that
                # point, but inform the developer about their mishap.
                warnings.warn(
                    "NoReverseMatch for %s while checking for pages allowed without 2FA"
                    % urlname
                )

        # User already has two-factor configured, do nothing.
        if get_adapter(request).has_2fa_enabled(request.user):
            return

        # The request required 2FA but it isn't configured!
        return self.on_require_2fa(request)



# ******************************************************************************
#
# app_settings.py:  application settings loaded from settings.py
#
# SPDX-License-Identifier: Apache-2.0
#
# django-allauth-2f2a, a 2fa adapter for django-allauth.
#
# ******************************************************************************
#
# Copyright 2016-2021 Víðir Valberg Guðmundsson and Percipient
# Networks, LLC.
# Copyright 2021 Jeremy A Gray <gray@flyquackswim.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.
#
# ******************************************************************************
#
"""Application settings loaded from settings.py."""

from allauth.account import app_settings as allauth_settings
from django.conf import settings

TEMPLATE_EXTENSION = getattr(
    settings,
    "ALLAUTH_2F2A_TEMPLATE_EXTENSION",
    allauth_settings.TEMPLATE_EXTENSION,
)

ALWAYS_REVEAL_BACKUP_TOKENS = bool(
    getattr(
        settings,
        "ALLAUTH_2F2A_ALWAYS_REVEAL_BACKUP_TOKENS",
        True,
    ),
)

QRCODE_TYPE = getattr(
    settings,
    "ALLAUTH_2F2A_QRCODE_TYPE",
    "data",
)

TWOFA_FORMS = getattr(
    settings,
    "ALLAUTH_2F2A_2FA_FORMS",
    {
        "authentication": "allauth_2f2a.forms.TOTPAuthenticateForm",
        "device": "allauth_2f2a.forms.TOTPDeviceForm",
        "remove": "allauth_2f2a.forms.TOTPDeviceRemoveForm",
    },
)



# ******************************************************************************
#
# setup.py:  django-allauth-2f2a build setup
#
# SPDX-License-Identifier: Apache-2.0
#
# django-allauth-2f2a, a 2fa adapter for django-allauth.
#
# ******************************************************************************
#
# Copyright 2016-2021 Víðir Valberg Guðmundsson and Percipient
# Networks, LLC.
# Copyright 2021 Jeremy A Gray <gray@flyquackswim.com>.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.
#
# ******************************************************************************
#
"""django-allauth-2f2a build setup."""

import codecs

from setuptools import find_packages
from setuptools import setup


def long_description():
    """Load README.rst for setup's long description."""
    with codecs.open("README.rst", encoding="utf8") as f:
        return f.read()


setup(
    name="django-allauth-2f2a",
    version="0.9.1",
    packages=find_packages(".", include=("allauth_2f2a", "allauth_2f2a.*")),
    include_package_data=True,
    install_requires=[
        "django>=2.2",
        "qrcode>=5.3",
        "django-allauth>=0.44",
        "django-otp>=1.0.0",
    ],
    author="Jeremy A Gray",
    author_email="gray@flyquackswim.com",
    description="Adds two factor authentication to django-allauth.",
    license="Apache 2.0",
    keywords=["otp", "auth", "two factor authentication", "allauth", "django", "2fa"],
    url="https://github.com/jeremyagray/django-allauth-2f2a",
    long_description=long_description(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Environment :: Web Environment",
        "Topic :: Internet",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.2",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: Apache Software License",
    ],
    python_requires=">=3.6",
)
